/*
 * Copyright 1993-2010 NVIDIA Corporation.  All rights reserved.
 *
 * NVIDIA Corporation and its licensors retain all intellectual property and 
 * proprietary rights in and to this software and related documentation. 
 * Any use, reproduction, disclosure, or distribution of this software 
 * and related documentation without an express license agreement from
 * NVIDIA Corporation is strictly prohibited.
 *
 * Please refer to the applicable NVIDIA end user license agreement (EULA) 
 * associated with this source code for terms and conditions that govern 
 * your use of this NVIDIA software.
 * 
 */

#include "cuda.h"
#include "../common/book.h"
#include "../common/cpu_bitmap.h"

#define DIM 1024

#define rnd( x ) (x * rand() / RAND_MAX)

#define INF     2e10f

#define SPHERES 20

// define the sphere structure
struct Sphere {
  float   r,b,g;
  float   radius;
  float   x,y,z;
  float hit( float ox, float oy, float *n ) 
  {
    float dx = ox - x;
    float dy = oy - y;
    if (dx*dx + dy*dy < radius*radius) {
      float dz = sqrtf( radius*radius - dx*dx - dy*dy );
      *n = dz / sqrtf( radius * radius );
      return dz + z;
    }
    return -INF;
  }
};


// calculate whether a ray intercept a sphere
// if so, where and what color
void rays( int x, int y, Sphere *s, unsigned char *ptr ) 
{
  float   ox = (x - DIM/2);
  float   oy = (y - DIM/2);    
  int offset = x + y * DIM;
    
  float   r=0, g=0, b=0;
  float   maxz = -INF;
  for(int i=0; i<SPHERES; i++) {
    float   n;
    float   t = s[i].hit( ox, oy, &n );
    if (t > maxz) {
      float fscale = n;
      r = s[i].r * fscale;
      g = s[i].g * fscale;
      b = s[i].b * fscale;
      maxz = t;
    }
  } 

  // record color at given piixel
  ptr[offset*4 + 0] = (int)(r * 255);
  ptr[offset*4 + 1] = (int)(g * 255);
  ptr[offset*4 + 2] = (int)(b * 255);
  ptr[offset*4 + 3] = 255;
}

// kernal scans the color image
void kernel( Sphere *s, unsigned char *ptr )
{
  int x, y;

  for ( y=0; y<DIM; y++ )
    for ( x=0; x<DIM; x++ ) 
      rays( x, y, s, ptr );
 }


int main( void ) 
{
  // create the start & stop time object
  cudaEvent_t start, stop;
  HANDLE_ERROR( cudaEventCreate( &start ) );
  HANDLE_ERROR( cudaEventCreate( &stop ) );

  // allocate memory for the Sphere dataset
  // and define the spheres in the dataset

  Sphere *s = (Sphere*)malloc( sizeof(Sphere) * SPHERES );
   
  for (int i=0; i<SPHERES; i++) {
    s[i].r = rnd( 1.0f );
    s[i].g = rnd( 1.0f );
    s[i].b = rnd( 1.0f );
    s[i].x = rnd( 1000.0f ) - 500;
    s[i].y = rnd( 1000.0f ) - 500;
    s[i].z = rnd( 1000.0f ) - 500;
    s[i].radius = rnd( 100.0f ) + 20;
  }

  HANDLE_ERROR( cudaEventRecord( start, 0 ) );

  // define the openGl bitmap 
  // in which to store image
  CPUBitmap bitmap( DIM, DIM );

  // alternate pointer
  unsigned char *ptr = bitmap.get_ptr();

  // kernal scan image & calculate
  // color of sphere if ray hits sphere
  kernel( s, ptr );

  // get stop time, and display the timing results
  HANDLE_ERROR( cudaEventRecord( stop, 0 ) );
  HANDLE_ERROR( cudaEventSynchronize( stop ) );
  float   elapsedTime;
  HANDLE_ERROR( cudaEventElapsedTime( &elapsedTime,
				      start, stop ) );
  printf( "Time to generate:  %3.1f ms\n", elapsedTime );

  // be good citizen bu cleaning up memory
  HANDLE_ERROR( cudaEventDestroy( start ) );
  HANDLE_ERROR( cudaEventDestroy( stop ) );
  
  bitmap.display_and_exit();
}

